/* emu.c: VMS pcap emulation network routines for SIMH
  ------------------------------------------------------------------------------
   Copyright (c) 2003-2004, Graham Burley

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the author.

  ------------------------------------------------------------------------------

  These routines emulate just enough PCAP functionality to support SIMH.

  Limitations:
   - requires a dedicated interface
   - only supports Ethernet packets (with 2 byte protocol type)

  The routines use the QIO interface to the LAN drivers. They work by
  inspecting the ethernet protocol type of outbound packets and starting
  ports as needed for that protocol. There is no attempt to fake the
  source address, when a port is started it attempts to change the
  physical address to that configured by SIMH. Effectitvely this limits
  the emulation to running on a dedicated interface.

  The routines do not enable Promiscous mode, however they do enable
  All Mulicast Mode on each port. There is no need for SIMH to filter
  packets.

  This has been tested with SIMH V3.1-0 running VAX/VMS 5.4-2 running
  DECnet V5.4-1, LAT V5.4-2, and UCX V1.3 (!). The only issue was that
  UCX does not send any IP packets on startup (only ARP) so some outbound
  traffic must be generated to start an IP port before SIMH can accept
  IP connections.


  ------------------------------------------------------------------------------

  Modification history:

  06-Sep-04  GB   Added pcap_datalink req w/3.2-3
  16-Jan-04  GB   Added pcap_find(free)alldevs
  12-Jan-04  GB   Tidied up for 3.1-0
  25-Oct-03  GB   Initial working version w/3.0-1

  ------------------------------------------------------------------------------
*/

#include <pcap.h>

#include <dcdef>
#include <devdef>
#include <dvidef>
#include <dvsdef>
#include <efndef>
#include <iodef>
#include <lib$routines>
#include <nmadef>			/* should be from SYS$LIB_C */
#include <ssdef>
#include <starlet>
#include <ucbdef>			/* should be from SYS$LIB_C */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#pragma member_alignment save
#pragma nomember_alignment
struct setmode_option {
	short	id;
	long	op;
};
struct setmode_address {
	short	id;
	short	len;
	short	mod;
	unsigned char address[6];
};
struct port_setmode {
	struct setmode_option opt[7];
	struct setmode_address pha;
};
struct iosb {
	short s;
	short len;
	long more;
};
struct s_item {
	short len;
	short code;
	long *buf;
	short *ret;
};
struct s_item1 {
	struct s_item item;
	long term;
	long retcode;
	short retlen;
};
struct s_item3 {
	struct s_item item[3];
	long term;
	long retcode[3];
	short retlen[3];
};
#pragma member_alignment restore

int find_port(pcap_t *, u_char *, int);
int start_port(pcap_t *, u_char *);
void q_read(pcap_t *, int);
void q_done(struct pcap_port *);
int check_dev(struct dsc$descriptor *, char *);
int open_dev(struct dsc$descriptor *, u_char *, char *);
void mod_decode( short len, char *buf, u_char *pha);

/*
 *
 */
pcap_t *pcap_open_live(char *device
	, int snaplen, int promisc, int to_ms, char *ebuf)
{
	int i;
	int s;
	pcap_t *p;
	short chan;

	p = malloc(sizeof(pcap_t));
	if (p == NULL) {
		sprintf(ebuf, "malloc");
		return(NULL);
	}
	memset(p, 0, sizeof(pcap_t));

	p->devnam_d.dsc$b_class = DSC$K_CLASS_S;
	p->devnam_d.dsc$w_length = strlen(device);
	p->devnam = malloc(p->devnam_d.dsc$w_length);
	if (p->devnam == NULL) {
		sprintf(ebuf, "malloc");
		free(p);
		return(NULL);
	}
	memcpy(p->devnam, device, p->devnam_d.dsc$w_length);
	p->devnam_d.dsc$a_pointer = p->devnam;

	s = check_dev(&p->devnam_d, ebuf);
	if (s != 1) {
		free(p->devnam);
		free(p);
		return(NULL);
	}

	s = open_dev(&p->devnam_d, &p->pha[0], ebuf);
	if (s != 1) {
		free(p->devnam);
		free(p);
		return(NULL);
	}

	return(p);
}

void pcap_close(pcap_t *p)
{
	int i;

	if (p == NULL) {
		return;
	}
	for (i=0; i<p->ports; i++) {
		if (p->port[i].chan != 0) {
			sys$dassgn(p->port[i].chan);
		}
	}
	if (p->devnam != NULL) {
		free(p->devnam);
	}
	free(p);
}

int pcap_sendpacket(pcap_t *p, u_char *packet, int len)
{
	int s;
	short x;
	struct iosb io;

	x = find_port(p, packet, len);
	if (x == -1) return(-1);

	s = sys$qiow(EFN$C_ENF, p->port[x].chan, IO$_WRITEVBLK
		, &io, 0, 0
		, &packet[14], len - 14, 0, 0, &packet[0], 0);

	if ((s != SS$_NORMAL) || (io.s != SS$_NORMAL)) {
		printf("pcap_emu: send: failed - %08x %04x %d\n"
			, s, io.s, io.len);
		return(-1);
	}

	return (io.len == len)?0:-1;
}

int pcap_dispatch(pcap_t *p, int cnt, pcap_handler callback, u_char *userdata)
{
	int i;
	int s;
	struct pcap_pkthdr pcap_header;
	int x;
	int r = 0;

	sys$setast(0);
	if (p->dq.len > 0) {
		x = p->dq.port[0];
		if ( p->port[x].io.s == SS$_NORMAL) {
			pcap_header.len = p->port[x].io.len + 14;
			callback(userdata, &pcap_header, p->port[x].buf);
			r = 1;
		} else {
			printf("pcap_emu: read: failed - %d %04x %d\n"
				, x, p->port[x].io.s, p->port[x].io.len);
			r = -1;
		}
		p->dq.len--;
		for (i=0;i<p->dq.len;i++) {
			p->dq.port[i] = p->dq.port[i+1];
		}
		q_read(p, x);
	}
	sys$setast(1);

	return(r);
}

int find_port(pcap_t *p, u_char *packet, int len)
{
	int i;
	short *pty;
	short x = -1;

	if (len <14) {
		return(-1);
	}

	pty = (short *) &packet[12];
	for (i=0; i<p->ports; i++) {
		if (p->port[i].pty == *pty) {
			x = i;
		}
	}
	if (x == -1) {
		x = start_port(p, packet);
		if (x == -1) {
			return(-1);
		}
		q_read(p, x);
	}

     /* check source address, we only allow change on start */

	if ( memcmp(&packet[6], &p->pha, 6) != 0 ) {
		printf("pcap_emu: send: invalid source address ");
		printf("%02x-%02x-%02x-%02x-%02x-%02x pty %04x\n"
			, packet[6], packet[7], packet[8]
			, packet[9], packet[10], packet[11], pty );
		return(-1);
	}

	return(x);
}

int start_port(pcap_t *p, u_char *packet)
{
	int s;
	long q[2];
	struct iosb io;
	struct port_setmode ps;
	short *pty;
	u_char *pha;
	short chan;
	int x;

	pty = (short *) &packet[12];

	if (p->ports < PCAP_MAX_PORTS) {
		x = p->ports;
	} else {
		printf("pcap_emu: start_port: out of ports\n");
		return(-1);
	}
	if (p->port[x].chan != 0) {
		printf("pcap_emu: start_port: non zero chan\n");
		return(-1);
	}

	s = sys$assign( &p->devnam_d, &chan, 0, 0);
	if (s != SS$_NORMAL) {
		printf("pcap_emu: start_port: failed - assign %08x\n", s);
		return(-1);
	}

	ps.opt[0].id = NMA$C_PCLI_FMT;	ps.opt[0].op = NMA$C_LINFM_ETH;
	ps.opt[1].id = NMA$C_PCLI_PAD;	ps.opt[1].op = NMA$C_STATE_OFF;
	ps.opt[2].id = NMA$C_PCLI_CCA;	ps.opt[2].op = NMA$C_STATE_ON;
	ps.opt[3].id = NMA$C_PCLI_MLT;	ps.opt[3].op = NMA$C_STATE_ON;
	ps.opt[4].id = NMA$C_PCLI_BFN;	ps.opt[4].op = 8;
	ps.opt[5].id = NMA$C_PCLI_BUS;	ps.opt[5].op = PCAP_MAX_PACKET;
	ps.opt[6].id = NMA$C_PCLI_PTY;	ps.opt[6].op = *pty;
	ps.pha.id = NMA$C_PCLI_PHA;
	ps.pha.len = 8;
	ps.pha.mod = NMA$C_LINMC_SET;
	memcpy(&ps.pha.address[0], &packet[6], 6);

	q[0] = sizeof(ps);
	q[1] = (long) &ps;
	s = sys$qiow(EFN$C_ENF, chan
		, IO$_SETMODE | IO$M_CTRL | IO$M_STARTUP
		, &io, 0, 0
		, 0, &q, 0, 0, 0, 0);
	if ((s != SS$_NORMAL) || (io.s != SS$_NORMAL)) {
		printf("pcap_emu: start_port: failed - %08x %04x %d\n"
			, s, io.s, io.more);
		sys$dassgn(chan);
		return(-1);
	}

	s = lib$get_ef(&p->port[x].efn);
	if (s != SS$_NORMAL) {
		printf("pcap_emu: start_port: failed - get_ef %08x\n", s);
		sys$dassgn(chan);
		return(-1);
	}
	p->port[x].index = x;
	p->port[x].dq = &p->dq;

	if ( memcmp(&ps.pha.address[0], &p->pha, 6) != 0 ) {
#ifdef PCAP_EMU_DEBUG
		printf("pcap_emu: pha: %02x-%02x-%02x-%02x-%02x-%02x\n"
			, ps.pha.address[0], ps.pha.address[1]
			, ps.pha.address[2], ps.pha.address[3]
			, ps.pha.address[4], ps.pha.address[5] );
#endif
		memcpy(&p->pha[0], &ps.pha.address[0], 6);
	}

	p->port[x].chan = chan;
	p->port[x].pty = *pty;
	p->ports++;

#ifdef PCAP_EMU_DEBUG
	printf("pcap_emu: start_port: %d chan %d pty %04x efn %d\n"
		, x, chan, *pty, p->port[x].efn);
#endif

	return(x);
}

void q_read(pcap_t *p, int x)
{
	int s;

	s = sys$qio( p->port[x].efn, p->port[x].chan, IO$_READVBLK
		, &p->port[x].io, &q_done, &p->port[x]
		, &p->port[x].buf[14], PCAP_MAX_PACKET - 14, 0, 0
		, &p->port[x].buf[0], 0);

	if (s != SS$_NORMAL) {
		printf("pcap_emu: q_read: failed port %d %08x\n", x, s);
	}
}

void q_done(struct pcap_port *p)	/* AST mode */
{
	p->dq->port[p->dq->len] = p->index;
	p->dq->len++;
}

int pcap_findalldevs(pcap_if_t **alldevsp, char *ebuf)
{
	int s;
	char buf[64];
	u_char pha[6];
	struct dsc$descriptor devnam;
	short devlen;
	long long ctx = 0;
	struct s_item1 devs;
	pcap_if_t *first = NULL;
	pcap_if_t *this = NULL;
	pcap_if_t *last = NULL;

	devs.item.len = 4;
	devs.item.code = DVS$_DEVCLASS;
	devs.item.buf = &devs.retcode;
	devs.item.ret = 0;
	devs.term = 0;
	devs.retcode = DC$_SCOM;

	devnam.dsc$b_class = DSC$K_CLASS_S;
	devnam.dsc$a_pointer = &buf[0];
	devnam.dsc$w_length = 64;

	*alldevsp = NULL;
again:
	s = sys$device_scan(&devnam, &devlen, 0, &devs , &ctx);
	if (s == SS$_NOMOREDEV) {
		*alldevsp = first;
		return(1);
	}
	if (s != SS$_NORMAL) {
		sprintf(ebuf, "device_scan %08x", s);
		pcap_freealldevs(first);
		return(-1);
	}

	s = check_dev(&devnam, ebuf);
	if (s == -1) {
		pcap_freealldevs(first);
		return(-1);
	}

	if (s == 1) {
		s = open_dev(&devnam, &pha[0], ebuf);
		if (s == -1) {
			pcap_freealldevs(first);
			return(-1);
		}
		this = malloc(sizeof(pcap_if_t));
		if (this == NULL) {
			sprintf(ebuf, "malloc");
			pcap_freealldevs(first);
			return(-1);
		}
		this->name = malloc(devnam.dsc$w_length + 1);
		this->description = malloc(18);
		if ((this->name == NULL) || (this->description == NULL)) {
			sprintf(ebuf, "malloc");
			free(this);
			pcap_freealldevs(first);
			return(-1);
		}
		memcpy(this->name, &buf[0], devlen);
		this->name[devlen] = '\0';
		sprintf(this->description, "%02X-%02X-%02X-%02X-%02X-%02X"
			, pha[0], pha[1], pha[2], pha[3], pha[4], pha[5]);
		this->flags = 0;

		if (last == NULL) {
			first = this;
		} else {
			last->next = this;
		}
		last = this;
	}

	goto again;
}

int pcap_freealldevs(pcap_if_t *p)
{
	pcap_if_t *next;

	while (p != NULL) {
		free(p->name);
		free(p->description);
		next = p->next;
		free(p);
		p = next;
	}

	return(1);
}

/*
 * Is it an Ethernet device?
 *
 *  IO Users 9.15 says DC$_SCOM with the following bits set
 *	DEV$M_AVL, DEV$M_IDV, DEV$M_NET, DEV$M_ODV
 *
 *  Logic says we must also limit to template devices
 *  TCPware creates IPA intefaces that are mailboxes
 *  Need someway to reject FDDI, etc interfaces?
 *
 *  ??? sense mode and check CSMACD and MAC adress ???
 *
 *
 */
int check_dev(struct dsc$descriptor *devnam, char *ebuf)
{
	int s;
	struct s_item3 dvi;
	struct iosb io;

	dvi.item[0].len = 4;
	dvi.item[0].code = DVI$_DEVCLASS;
	dvi.item[0].buf = &dvi.retcode[0];
	dvi.item[0].ret = &dvi.retlen[0];
	dvi.item[1].len = 4;
	dvi.item[1].code = DVI$_DEVCHAR;
	dvi.item[1].buf = &dvi.retcode[1];
	dvi.item[1].ret = &dvi.retlen[1];
	dvi.item[2].len = 4;
	dvi.item[2].code = DVI$_STS;
	dvi.item[2].buf = &dvi.retcode[2];
	dvi.item[2].ret = &dvi.retlen[2];
	dvi.term = 0;

	s = sys$getdviw(EFN$C_ENF, 0, devnam, &dvi, &io, 0, 0, 0);
	if (s != SS$_NORMAL) {
		sprintf(ebuf, "getdvi %08x", s);
		return(-1);
	}

	if ( (dvi.retcode[0] == DC$_SCOM)
	  && (dvi.retcode[1] & DEV$M_AVL) && (dvi.retcode[1] & DEV$M_IDV)
	  && (dvi.retcode[1] & DEV$M_NET) && (dvi.retcode[1] & DEV$M_ODV)
	  && !(dvi.retcode[1] & DEV$M_MBX)
	  && (dvi.retcode[2] & UCB$M_TEMPLATE) ) {
		return(1);
	} else {
		sprintf(ebuf, "not an Ethernet device");
		return(0);
	}
}

/*
 * See if we can open device, start port using custumer reserved protocol
 * save pha for possible use
 *
 */
int open_dev(struct dsc$descriptor *devnam, u_char *pha, char *ebuf)
{
	int s;
	long q[2];
	struct iosb io;
	struct port_setmode ps;
	short pty = 0x0660;	/* 6006 Customer Use */
	short chan;
	char mod[250];		/* sensemode buffer */

	s = sys$assign( devnam, &chan, 0, 0);
	if (s != SS$_NORMAL) {
		sprintf(ebuf, "assign %08x\n", s);
		return(-1);
	}

	ps.opt[0].id = NMA$C_PCLI_FMT;	ps.opt[0].op = NMA$C_LINFM_ETH;
	ps.opt[1].id = NMA$C_PCLI_PAD;	ps.opt[1].op = NMA$C_STATE_OFF;
	ps.opt[2].id = NMA$C_PCLI_CCA;	ps.opt[2].op = NMA$C_STATE_ON;
	ps.opt[3].id = NMA$C_PCLI_MLT;	ps.opt[3].op = NMA$C_STATE_ON;
	ps.opt[4].id = NMA$C_PCLI_BFN;	ps.opt[4].op = 8;
	ps.opt[5].id = NMA$C_PCLI_BUS;	ps.opt[5].op = PCAP_MAX_PACKET;
	ps.opt[6].id = NMA$C_PCLI_PTY;	ps.opt[6].op = pty;
	/* not setting pha */
	q[0] = 7 * 6;
	q[1] = (long) &ps;
	s = sys$qiow(EFN$C_ENF, chan
		, IO$_SETMODE | IO$M_CTRL | IO$M_STARTUP
		, &io, 0, 0
		, 0, &q, 0, 0, 0, 0);
	if ((s != SS$_NORMAL) || (io.s != SS$_NORMAL)) {
		sprintf(ebuf, "startup failed - %08x %04x %d\n"
			, s, io.s, io.more);
		sys$dassgn(chan);
		return(-1);
	}

  /* Sensemode to get physical address */

	q[0] = 250;
	q[1] = (long) &mod[0];

	s = sys$qiow(EFN$C_ENF, chan, IO$_SENSEMODE | IO$M_CTRL
		, &io, 0, 0
		, 0,&q, 0, 0, 0, 0);
	if ((s != SS$_NORMAL) || (io.s != SS$_NORMAL)) {
		sprintf(ebuf, "sensemode failed - %08x %04x %d\n"
			, s, io.s, io.more);
		sys$dassgn(chan);
		return(-1);
	}
	mod_decode( io.len, &mod[0], pha);

  /* We're done */

	sys$dassgn(chan);
	return(1);
}

void mod_decode( short len, char *buf, u_char *pha)
{
	short *p;
	short i=0;
	short id, typ;
	long *val;

	while (i < len) {
		p = (short *) &buf[i];
		typ = (0x1000 & *p) >> 12;
		id  = 0x0fff & *p;
		i += 2;
		p++;
		if (typ == 1) {				/* counted string */
			if (id == NMA$C_PCLI_PHA) {
				memcpy(pha, &buf[i+2], 6);
			}
			i = i + 2 + *p;
		} else {				/* longword */
			val = (long *) &buf[i];
			i += 4;
		}
	}
}

int pcap_datalink(pcap_t *p)
{
	return(DLT_EN10MB);
}
